from typing import List, Tuple, Dict, Union
import pandas as pd
from sklearn.tree import DecisionTreeClassifier
from .utils import extract_rules, compress_rule

def get_rules_decision_tree(
    X_delta: pd.DataFrame,
    y_delta: pd.DataFrame,
    params: Dict[str, Union[int, float, str]] = {
        "max_depth": 4,
        "criterion": "gini",
        "min_samples_leaf": 10,
        "min_impurity_decrease": 0,
    },
    class_diff: int = 1
) -> Tuple[DecisionTreeClassifier, List[str]]:
    """
    Extract rules from a decision tree.

    Args:
        X_delta (pd.DataFrame): Input data for deltamodels
        y_delta (pd.DataFrame): Disagreeing labels
        params (dict, optional): Decision tree parameters. Defaults to {"max_depth":4, "criterion":"gini", "min_samples_leaf":10,"min_impurity_decrease": 0}.
        class_diff (int, optional): Class to consider as disagreeing. Defaults to 1.

    Returns:
        Tuple[DecisionTreeClassifier, List[str]]: Trained deltamodel and list of segments where models make different predictions
    """
    max_depth = params["max_depth"]
    criterion = params["criterion"]
    min_samples_leaf = params["min_samples_leaf"]
    min_impurity_decrease = params["min_impurity_decrease"]

    delta = DecisionTreeClassifier(
        max_depth=max_depth,
        criterion=criterion,
        min_samples_leaf=min_samples_leaf,
        min_impurity_decrease=min_impurity_decrease
    )
    delta.fit(X_delta, y_delta)
    raw_rules = extract_rules(delta, feature_names=X_delta.columns, class_names=["0", "1"], class_diff=class_diff)
    segments = [compress_rule(rule) for rule in raw_rules]
    return delta, segments



from sklearn.tree import _tree
import numpy as np
from typing import List, Optional

def extract_rules(
    tree: _tree.Tree,
    feature_names: List[str],
    class_names: Optional[List[str]],
    class_diff: int
) -> List[str]:
    """
    Extract rules from a decision tree.

    Args:
        tree (_tree.Tree): Model to extract rules from
        feature_names (List[str]): Feature names
        class_names (List[str], optional): Class names
        class_diff (int): Class to consider as disagreeing

    Returns:
        List[str]: Extracted rules
    """
    tree_ = tree.tree_
    feature_name = [
        feature_names[i] if i != _tree.TREE_UNDEFINED else "undefined!"
        for i in tree_.feature
    ]

    paths = []
    path = []
    
    def recurse(node, path, paths):
        
        if tree_.feature[node] != _tree.TREE_UNDEFINED:
            name = feature_name[node]
            threshold = tree_.threshold[node]
            p1, p2 = list(path), list(path)
            p1 += [f"({name} <= {np.round(threshold, 3)})"]
            recurse(tree_.children_left[node], p1, paths)
            p2 += [f"({name} > {np.round(threshold, 3)})"]
            recurse(tree_.children_right[node], p2, paths)
        else:
            path += [(tree_.value[node], tree_.n_node_samples[node])]
            paths += [path]
            
    recurse(0, path, paths)

    # sort by samples count
    samples_count = [p[-1][1] for p in paths]
    ii = list(np.argsort(samples_count))
    paths = [paths[i] for i in reversed(ii)]
    
    rules = []
    for path in paths:
        rule = "if "
        
        for p in path[:-1]:
            if rule != "if ":
                rule += " and "
            rule += str(p)
        rule += " then "
        if class_names is None:
            rule += "response: "+str(np.round(path[-1][0][0][0],3))
        else:
            classes = path[-1][0][0]
            l = np.argmax(classes)
            rule += f"class: {class_names[l]} (Probability: {np.round(100.0*classes[l]/np.sum(classes),2)}%)"
        rule += f" | based on {path[-1][1]:,} samples"
        rules += [rule]
    
    delta_rules = []
    for rule in rules:
        if "then class: " + str(class_diff) in rule:
            delta_rules.append(rule)
    return delta_rules

def compress_rule(rule: str) -> str:
    """
    Compresses a rule extracted from a decision tree.

    Args:
        rule (str): Rule extracted from the decision tree

    Returns:
        str: Compressed rule
    """
    def parse_condition(condition):
        condition = condition.replace('(', '').replace(')', '')
        if '<=' in condition:
            feature, value = condition.split(' <= ')
            operator = '<='
        elif '>=' in condition:
            feature, value = condition.split(' >= ')
            operator = '>='
        elif '<' in condition:
            feature, value = condition.split(' < ')
            operator = '<'
        elif '>' in condition:
            feature, value = condition.split(' > ')
            operator = '>'
        else:
            print("Condition :",condition)
            raise()
            return "", "", 0
        return feature.strip(), operator, float(value.strip())
    condition_list = rule.split('if ')[1].split(' then ')[0].split(' and ')
    second_part = rule.split('if ')[1].split(' then ')[1]
    all_conds = []
    for condition in condition_list:
        all_conds.append(parse_condition(condition))
    new_conds = {}
    for ext_cond in all_conds:
        feature, operator, value = ext_cond
        key  = (feature,operator)
        if key not in new_conds:
            new_conds[key] = value
        else:
            if (operator in ["<", "<="] and value <= new_conds[key]) or (operator in [">", ">="] and value >= new_conds[key]):
                new_conds[key] = value

    substring = "if "
    i = 1
    for cond in new_conds:

        element = "(" + cond[0] +" "+ cond[1] +" "+  str(new_conds[cond]) + ")"
        substring += element
        if i == len(new_conds):
            substring += " then " + second_part
        else:
            substring += " and "
        i += 1
    return(substring)


from setuptools import setup

setup(
    name='deltaxplainer',
    version='1.1.2',
    packages=['deltaxplainer'],
    install_requires=[
    'numpy>=1.20',
    'scikit-learn>=0.24',
    'pandas>=1.3',
    'pytest'
    ],
    author='Adam Rida',
    description='A package for DeltaXplainer model implemented from the paper https://arxiv.org/pdf/2309.17095.pdf',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
)



from typing import Optional, Dict, List, Union
import pandas as pd
from sklearn.base import BaseEstimator, RegressorMixin, ClassifierMixin
from .decision_tree import get_rules_decision_tree

class DeltaXplainer(BaseEstimator, RegressorMixin):
    def __init__(self) -> None:
        self.X_delta: Optional[pd.DataFrame] = None
        self.segments: List[str] = []
        self.model = None
        self.f = None
        self.g = None

    def fit(
        self,
        Xf: pd.DataFrame,
        f: ClassifierMixin,
        g: ClassifierMixin,
        Xg: Optional[pd.DataFrame] = None,
        yg: Optional[pd.Series] = None,
        params: Dict[str, Union[int, float, str]] = {
            "max_depth": 8,
            "criterion": "gini",
            "min_samples_leaf": 1,
            "min_impurity_decrease": 0,
        },
        class_diff: int = 1,
        
    ) -> 'DeltaXplainer':
        """
        Fits the DeltaModel as a Decision Tree and generates decision rules

        Args:
            Xf (pd.DataFrame): Feature set of the first group
            f (ClassifierMixin) : Model f - First model to compare
            g (ClassifierMixin) : Model g - Second model to compare
            Xg (pd.DataFrame, optional): Feature set of the second group. Defaults to None.
            yg (pd.Series, optional): Target variable of the second group. Defaults to None.
            params (dict, optional): Parameters for the decision tree. Defaults to {"max_depth": 4, "criterion": "gini", "min_samples_leaf": 10, "min_impurity_decrease": 0}.
            class_diff (int, optional): Class difference. Defaults to 1.

        Returns:
            DeltaXplainer: Fitted DeltaXplainer instance
        """
        
        self.f = f
        self.g = g
        if Xg is not None and yg is not None:
            self.X_delta = pd.concat([Xf, Xg])
        else:
            self.X_delta = Xf
            
        self.y_delta=(self.f.predict(self.X_delta) != self.g.predict(self.X_delta)).astype(int)
        
        self.model,self.segments = get_rules_decision_tree(self.X_delta, self.y_delta, params=params, class_diff=class_diff)
        return self

    def predict(self, X: pd.DataFrame) -> Union[pd.Series, None]:
        """
        Predicts the target variable for the given features

        Args:
            X (pd.DataFrame): Feature set for prediction

        Returns:
            Union[pd.Series, None]: Predicted target variable
        """
        if self.model is None:
            raise ValueError("Model not fitted yet. Call fit method first.")
        return self.model.predict(X)
    
    def get_segments(self) -> List[str]:
        """
        Returns the generated segments/rules

        Returns:
            List[str]: List of segments/rules
        """
        return self.segments
